'''
Unit test for LogStore class
'''

import logging
from dataclasses import dataclass

import pytest
from ap_utilities.logging.log_store import LogStore

# --------------------------------
@dataclass
class Data:
    '''
    Class used to store shared data
    '''
    l_backend = ['logging', 'logzero']
    l_level   = [10, 20, 30, 40, 50]
# --------------------------------
@pytest.mark.parametrize('backend', Data.l_backend)
def test_show(backend : str):
    '''
    Test for show_loggers
    '''
    LogStore.backend = backend

    name_war = f'show_warning_{backend}'
    name_def = f'show_default_{backend}'

    LogStore.set_level(name_war, logging.WARNING)

    LogStore.add_logger(name_war)
    LogStore.add_logger(name_def)

    LogStore.show_loggers()
# --------------------------------
@pytest.mark.parametrize('backend', Data.l_backend)
def test_messages(backend : str):
    '''
    Tests each level
    '''
    LogStore.backend = backend

    name = f'messages_{backend}'
    log = LogStore.add_logger(name)
    LogStore.set_level(name, 10)

    log.debug('debug')
    log.info('info')
    log.warning('warning')
    log.error('error')
    log.critical('critical')
# --------------------------------
@pytest.mark.parametrize('backend', Data.l_backend)
@pytest.mark.parametrize('level'  , Data.l_level)
def test_level(backend : str, level : int):
    '''
    Test for level setting
    '''
    LogStore.backend = backend

    name = f'level_{backend}_{level}'

    LogStore.add_logger(name)
    LogStore.set_level(name, level)

    LogStore.show_loggers()
# --------------------------------
@pytest.mark.parametrize('level', Data.l_level)
def test_logzero(level : int):
    '''
    Tests logzero
    '''
    LogStore.backend = 'logzero'

    name = f'logzero_{level}'
    log  = LogStore.add_logger(name)
    LogStore.set_level(name, level)

    print(30 * '-')
    print(f'Level: {level}')
    print(30 * '-')
    log.debug('debug')
    log.info('info')
    log.warning('warning')
    log.error('error')
    log.critical('critical')
    print(30 * '-')



'''
Script meant to invert the evt_name.yaml file
such that the keys are the nicknames and the values are the event types
'''
import argparse
from importlib.resources import files
import yaml

import ap_utilities.decays.utilities as aput

from ap_utilities.logging.log_store import LogStore

log = LogStore.add_logger('ap_utilities:make_name_evt')
# ------------------------------
def _get_data() -> dict[str,str]:
    cfg_path = files('ap_utilities_data').joinpath('evt_name.yaml')
    cfg_path = str(cfg_path)
    with open(cfg_path, encoding='utf-8') as ifile:
        d_evt_name = yaml.safe_load(ifile)

    return d_evt_name
# ------------------------------
def _invert_dict(d_evt_name : dict[str,str]) -> dict[str,str]:
    d_name_evt = {}
    for evt, name in d_evt_name.items():
        name             = aput.format_nickname(nickname=name, style= 'safe_1')
        d_name_evt[name] = evt

    return d_name_evt
# ------------------------------
def _parse_arguments():
    parser = argparse.ArgumentParser(description='Script used to invert evttype/nickname dictionary and save to YAML')
    _      = parser.parse_args()
# ------------------------------
def main():
    '''
    Script starts here
    '''
    _parse_arguments()

    d_evt_name = _get_data()
    d_name_evt = _invert_dict(d_evt_name)

    ofile_path = files('ap_utilities_data').joinpath('name_evt.yaml')
    ofile_path = str(ofile_path)

    log.info(f'Saving to: {ofile_path}')
    with open(ofile_path, 'w', encoding='utf-8') as ofile:
        yaml.safe_dump(d_name_evt, ofile)
# ------------------------------
if __name__ == '__main__':
    main()



'''
Script meant to read information from files in

https://gitlab.cern.ch/lhcb-datapkg/Gen/DecFiles

and store it in current project as data
'''
import os
import glob
from dataclasses           import dataclass
from importlib.resources   import files

import tqdm
import yaml
from ap_utilities.logging.log_store import LogStore

log=LogStore.add_logger('ap_utilities_scripts:update_decinfo')
# ------------------------------
@dataclass
class Data:
    '''
    Class used to store shared data
    '''
    dec_path : str
# ------------------------------
def _setup() -> None:
    if 'DECPATH' not in os.environ:
        raise ValueError('DECPATH, path to root of DecFiles, not found')

    Data.dec_path = os.environ['DECPATH']
# ------------------------------
def _line_from_list(file_path : str, contains : str, l_line : list[str]) -> str:
    l_value = [ line for line in l_line if contains in line ]

    if len(l_value) == 0:
        log.warning(f'Could not extract {contains} line in: {file_path}')
        return 'not_found'

    return l_value[0]
# ------------------------------
def _val_from_line(file_path : str, line : str) -> str:
    if line == 'not_found':
        return line

    l_part = line.split(':')
    if len(l_part) != 2:
        raise ValueError(f'In {file_path}, expected two elements separated by colon, found: \"{line}\"')

    part   = l_part[1]
    part   = part.rstrip().lstrip()

    return part
# ------------------------------
def _get_evt_info(kind : str, file_path : str) -> tuple[str,str]:
    with open(file_path, encoding='utf-8') as ifile:
        l_line = ifile.read().splitlines()

    evt_line = _line_from_list(file_path, 'EventType', l_line)
    kin_line = _line_from_list(file_path,       kind , l_line)

    evt_type = _val_from_line(file_path, evt_line)
    info     = _val_from_line(file_path, kin_line)

    return evt_type, info
# ------------------------------
def _read_info(kind : str) -> dict[str,str]:
    dec_file_wc = f'{Data.dec_path}/dkfiles/*.dec'
    l_dec_file  = glob.glob(dec_file_wc)
    nfiles      = len(l_dec_file)
    if nfiles == 0:
        raise ValueError(f'No dec file foudn in {dec_file_wc}')

    log.info(f'Found {nfiles} decay files')

    l_evt_info = [ _get_evt_info(kind, file_path) for file_path in tqdm.tqdm(l_dec_file, ascii=' -') ]
    d_evt_info = _dict_from_tup_list(l_evt_info)

    return d_evt_info
# ------------------------------
def _dict_from_tup_list(l_evt_name : list[tuple[str,str]]) -> dict[str,str]:
    d_res = {}
    for key, val in l_evt_name:
        if key in d_res:
            old_val = d_res[key]
            log.warning(f'Key {key} with value {old_val} already found, overriding with {val}')

        d_res[key] = val

    return d_res
# ------------------------------
def _dump_info(name : str, d_evt_info: dict[str,str]) -> None:
    yaml_path = files('ap_utilities_data').joinpath(f'{name}.yaml')
    yaml_path = str(yaml_path)

    log.info(f'Saving to: {yaml_path}')
    with open(yaml_path, 'w', encoding='utf-8') as ofile:
        yaml.dump(d_evt_info, ofile, width=500)
# ------------------------------
def main():
    '''
    Script starts here
    '''
    _setup()
    d_evt_name = _read_info(kind =   'NickName')
    _dump_info('evt_name', d_evt_name)

    d_evt_dec  = _read_info(kind = 'Descriptor')
    _dump_info('evt_dec' , d_evt_dec )
# ------------------------------
if __name__ == '__main__':
    main()



'''
Script used to build decay fields from YAML file storing event type -> decay correspondence
'''
import re
import argparse
from typing                         import Union
from importlib.resources            import files

import yaml

import ap_utilities.decays.utilities as aput
from ap_utilities.logging.log_store import LogStore

log = LogStore.add_logger('ap_utilities:make_fields')
# ---------------------------
class Data:
    '''
    Class storing shared data
    '''
    dec_head_rgx = r'\s+(\w+):'

    l_skip_type= [
            '12952000',
            '11453001',
            '13454001',
            '15454101',
            '12442001',
            '11442001',
            '13442001',
            '15444001',
            '15442001',
            '12873002',
            '12425011',
            ]

    d_repl_sym = {
            'cc'        :      'CC',
            '->'        :     '==>',
            }

    d_repl_par = {
            'psi(2S)'    :    'psi_2S_',
            'psi(1S)'    :    'psi_1S_',
            'K*(892)'    :    'K*_892_',
            'phi(1020)'  :  'phi_1020_',
            'K_1(1270)'  :  'K_1_1270_',
            'K_2*(1430)' : 'K_2*_1430_',
            }

    d_repl_spa = {
            '('        :     ' ( ',
            ')'        :     ' ) ',
            '['        :     ' [ ',
            ']'        :     ' ] ',
            }

    l_event_type : list[str]
    d_decay      : dict[str,str]

    d_nicknames = {
            'pi0' : 'pi0',
            'pi+' : 'pip',
            'pi-' : 'pim',
            'X'   : 'X',

            'K+'  : 'Kp',
            'K-'  : 'Km',

            'e+'  : 'Ep',
            'e-'  : 'Em',

            'mu+' : 'Mp',
            'mu-' : 'Mm',

            'tau+': 'taup',
            'tau-': 'taum',

            'p+'  : 'Pp',
            'p~-' : 'Pm',
            'K_S0': 'KS',

            'D-'     : 'D',
            'D0'     : 'D',
            'D0~'    : 'D',
            'D~0'    : 'D',
            'D_s-'   : 'D',
            'D*_s-'  : 'D',
            'nu_tau' : 'nu',
            'nu_mu'  : 'nu',
            'nu_e'   : 'nu',
            'nu_tau~': 'nu',
            'nu_mu~' : 'nu',
            'nu_e~'  : 'nu',

            'B+'  : 'Bu',
            'B-'  : 'Bu',
            'B0'  : 'Bd',
            'X0'  :  'X',
            'B_s0': 'Bs',
            'phi' : 'phi',
            'eta' : 'eta',

            'K_2*(1430)+'    : 'K2',
            'Beauty'         : 'B',
            'K_1+'           : 'K1',
            'K_1(1270)0'     : 'K1',
            'K_1(1270)+'     : 'K1',
            'phi(1020)'      : 'phi',
            'gamma'          : 'gm',
            'K*(892)0'       : 'Kst',
            'K*(892)+'       : 'Kstp',
            'J/psi(1S)'      : 'Jpsi',
            'psi(2S)'        : 'psi2S',
            'Lambda_b0'      : 'Lb',
            'Lambda0'        : 'Lz',
            'Lambda~0'       : 'Lz',
            'Lambda_c-'      : 'Lc',
            'Lambda_c~-'     : 'Lc',
            }
# ---------------------------
def _load_decays() -> None:
    dec_path = files('ap_utilities_data').joinpath('evt_dec.yaml')
    dec_path = str(dec_path)
    with open(dec_path, encoding='utf-8') as ifile:
        Data.d_decay = yaml.safe_load(ifile)
# ---------------------------
def _parse_args() -> None:
    parser = argparse.ArgumentParser(description='Used to perform several operations on TCKs')
    parser.add_argument('-i', '--input'   , type=str, help='Path to textfile with event types')
    parser.add_argument('-l', '--log_lvl' , type=int, help='Logging level', choices=[10,20,30], default=20)
    args = parser.parse_args()

    input_path = args.input
    with open(input_path, encoding='utf-8') as ifile:
        Data.l_event_type = ifile.read().splitlines()

    LogStore.set_level('ap_utilities:make_fields', args.log_lvl)
# ---------------------------
def _reformat_decay(decay : str) -> str:
    # Symbol renaming needed, e.g. -> ==>, cc -> CC
    for org, new in Data.d_repl_sym.items():
        decay = decay.replace(org, new)

    # Need to make special substrings into underscored ones
    # e.g. J/psi(1S) -> J/psi_1S_
    for org, new in Data.d_repl_par.items():
        decay = decay.replace(org, new)

    # Add spaces to parentheses and brackets
    for org, new in Data.d_repl_spa.items():
        decay = decay.replace(org, new)

    # Underscores are part of neutrino names
    # Particles otherwise only use letters
    # Numbers should be excluded, due to anti-D0 -> D~0
    decay = re.sub(r'anti-([a-zA-Z,_]+)', r'\1~', decay)

    return decay
# ---------------------------
def _reformat_back_decay(decay : str) -> str:
    # Put back special characters original naming
    for org, new in Data.d_repl_par.items():
        decay = decay.replace(new, org)

    # Decay cannot have space here, other spaces are allowed
    decay = decay.replace('] CC', ']CC')
    decay = decay.replace('[ '  ,   '[')
    decay = decay.replace('[  ' ,   '[')
    decay = decay.replace('  ]' ,   ']')
    decay = decay.replace('  [' ,   '[')
    decay = decay.replace(' ['  ,   '[')
    decay = decay.replace('(  ' ,   '(')
    decay = decay.replace('  )' ,   ')')

    return decay
# ---------------------------
def _replace_back(part : str) -> str:
    for org, new in Data.d_repl_par.items():
        if new in part:
            part = part.replace(new, org)

    return part
# ---------------------------
def _particles_from_decay(decay : str) -> list[str]:
    l_repl = list(Data.d_repl_sym.values())
    l_repl+= list(Data.d_repl_spa.values())
    l_repl = [ repl.replace(' ', '') for repl in l_repl ]

    l_part = decay.split(' ')
    l_part = [ part for part in l_part if part not in l_repl ]
    l_part = [ part for part in l_part if part != ''         ]
    l_part = [ _replace_back(part) for part in l_part ]

    # Anti-neutrinos and neutrinos will use nu(_index) branch names
    # Need this to make sure neutrinos appear as repeated
    l_part = [ part.rstrip('~') for part in l_part ]

    log.debug(f'Found particles: {l_part}')

    return l_part
# ---------------------------
def _skip_decay(event_type : str, decay : str) -> bool:
    if event_type in Data.l_skip_type:
        log.debug(f'Skipping decay: {decay}')
        return True

    if '{,gamma}' in decay:
        log.warning(f'Skipping {event_type} decay: {decay}')
        return True

    if 'nos' in decay:
        nickname = aput.read_decay_name(event_type=event_type, style= 'safe_1')
        log.warning('Skipping decay:')
        log.info(f'{"":<4}{nickname}')
        log.info(f'{"":<4}{event_type}')
        log.info(f'{"":<4}{decay}')

        return True

    return False
# ---------------------------
def _remove_index(particle : str) -> tuple[str,int]:
    '''
    Takes string representing particle and index of occurence
    Returns particle name and index in a tuple
    '''
    mtch = re.match(r'(.*)_(\d+)$', particle)
    if not mtch:
        return particle, 1

    particle = mtch.group(1)
    npar     = mtch.group(2)
    npar     = int(npar)

    return particle, npar
# ---------------------------
def _get_hatted_decay( particle : str, i_par : int, decay : str) -> str:
    decay = decay.replace(' '   , '  ')
    decay = decay.replace('   ' , '  ')
    decay = decay.replace('    ', '  ')

    if i_par == 0:
        return decay

    particle, ipar = _remove_index(particle)

    decay = _replace_nth_particle(decay, particle, ipar)

    return decay
# ---------------------------
def _move_hat(decay : str) -> str:
    '''
    Up to this point, hats are just behind every particle, including intermediates
    This function will move the hats before parenthesis
    '''
    org_decay = decay
    ihat      = decay.index('^')
    decay     = decay[:ihat]
    decay     = decay.rstrip()
    elm       = decay[-1]

    # Is first non-empty char before hat an opening parenthesis?
    # If not return
    if elm != '(':
        return org_decay

    # Otherwise remove hat from where it is and move it to right place
    ipar      = len(decay) - 1
    org_decay = org_decay.replace('^', ' ')
    decay     = org_decay[:ipar - 1] + '^' + org_decay[ipar:]

    return decay
# ---------------------------
def _replace_nth_particle(decay : str, particle:str, ipar:int) -> str:
    src    = f' {particle}'
    tgt    = f'^{particle}'

    l_part = decay.split(src)
    npart  = len(l_part)
    if npart == 1:
        raise ValueError(f'Cannot find {particle} in {decay}')

    if npart == 2:
        decay = decay.replace(src, tgt)
    else:
        decay = src.join(l_part[:ipar]) + tgt + src.join(l_part[ipar:])

    decay = _move_hat(decay)

    return decay
# ---------------------------
def _rename_repeated(l_par : list[str]) -> list[str]:
    '''
    Takes names of particles
    Returns names of particles, if particles appear more than once, append _x to name
    '''
    d_par_freq = {}
    for par in l_par:
        if par not in d_par_freq:
            d_par_freq[par] = 1
            continue

        d_par_freq[par]+= 1

    l_par_renamed = []
    for par, freq in d_par_freq.items():
        if freq == 1:
            l_par_renamed.append(par)
        else:
            l_par_renamed += [ f'{par}_{i_par}' for i_par in range(1, freq + 1) ]

    return l_par_renamed
# ---------------------------
def _nickname_from_particle(name : str, event_type : str, decname : str) -> str:
    name, ipar = _remove_index(name)
    # Nicknames will be the same for particles and antiparticles
    name       = name.replace('anti-', '')

    if name not in Data.d_nicknames:
        log.warning(f'Nickname for {name} not found in {decname}/{event_type}')
        return name

    nick = Data.d_nicknames[name]
    if ipar > 1:
        nick = f'{nick}_{ipar}'

    return nick
# ---------------------------
def _fix_beauty(decay : str, event_type : str) -> str:
    if 'Beauty' not in decay:
        return decay

    if event_type == '11102453':
        bname = 'B0'
    else:
        log.warning(f'Cannot identify B meson type for {event_type}')
        bname = 'Beauty'

    decay = decay.replace('Beauty', bname)

    return decay
# ---------------------------
def _fix_phi(decay : str) -> str:
    rgx   = r'phi(?!\s*\(\s*1020\s*\)\s*)'
    decay = re.sub(rgx, 'phi(1020)', decay)

    return decay
# ---------------------------
def _fix_names(decay : str, event_type : str) -> str:
    '''
    Decay field in decay files is not properly written, need to fix here, before using decay
    '''
    decay = decay.replace('K_1+' ,  'K_1(1270)+')
    decay = decay.replace('K*+'  ,    'K*(892)+')
    decay = decay.replace('K*0'  ,    'K*(892)0')
    decay = decay.replace('D_s*' ,        'D*_s')
    decay = decay.replace('My_'  ,            '')
    decay = _fix_phi(decay)
    decay = _fix_beauty(decay, event_type)

    return decay
# ---------------------------
def _get_decay(event_type : str, decname : str) -> Union[None,dict[str,str]]:
    decay = Data.d_decay[event_type]
    decay = _fix_names(decay, event_type)

    if _skip_decay(event_type, decay):
        return None

    decay = _reformat_decay(decay)
    l_par = _particles_from_decay(decay)
    l_par = _rename_repeated(l_par)
    decay = _reformat_back_decay(decay)
    decay = _check_unhatted_decay(decay, event_type)

    d_dec = {}
    for i_par, par in enumerate(l_par):
        nickname        = _nickname_from_particle(par, event_type, decname)
        d_dec[nickname] = _get_hatted_decay(par, i_par, decay)

    return d_dec
# ---------------------------
def _check_closed(decay : str, left : str, right : str) -> None:
    nlft = decay.count( left)
    nrgt = decay.count(right)

    if nlft != nrgt:
        log.error(f'Failed closure {left}{right} in {decay}')
# ---------------------------
def _check_unhatted_decay(decay : str, event_type : str) -> str:
    '''
    Final check of decay, will apply fixes if needed
    '''
    if not decay.endswith('CC'):
        log.warning(f'Decay {decay}/{event_type} had no conjugate adding it')
        return f'{decay}CC'

    _check_closed(decay, '(', ')')
    _check_closed(decay, '[', ']')

    return decay
# ---------------------------
def _get_decays() -> dict[str, dict[str,str]]:
    d_decay = {}
    for event_type in Data.l_event_type:
        decname = aput.read_decay_name(event_type=event_type, style= 'safe_1')
        d_tmp   = _get_decay(event_type, decname)
        if d_tmp is None:
            continue

        d_decay[decname] = d_tmp

    return d_decay
# ---------------------------
def _remove_ending_spaces(line : str) -> str:
    '''
    In the lines with a quote at the end, do:

    ']   ' -> ']'
    '''
    if '\'' not in line:
        return line

    line = line.rstrip()
    line = line.rstrip('\'')
    line = line.rstrip()

    return f'{line}' + '\''
# ---------------------------
def _format_yaml_line(line : str) -> str:
    line = _remove_ending_spaces(line)

    mtch = re.match(Data.dec_head_rgx, line)
    if not mtch:
        return line

    head     = mtch.group(1)
    head_pad = f'{head:5}'

    line = line.replace(f'{head}:', f'{head_pad}:')

    return line
# ---------------------------
def _save_decays(path : str, d_decay : dict[str,dict[str,str]]) -> None:
    with open(path, 'w', encoding='utf-8') as ofile:
        yaml.safe_dump(d_decay, ofile, width=200)

    with open(path, encoding='utf-8') as ifile:
        l_line = ifile.read().splitlines()

    l_line = [ _format_yaml_line(line) for line in l_line ]

    text = '\n'.join(l_line)
    with open(path, 'w', encoding='utf-8') as ofile:
        ofile.write(text)
# ---------------------------
def main():
    '''
    Script starts here
    '''
    _parse_args()
    _load_decays()

    d_decay = _get_decays()
    _save_decays('decays.yaml', d_decay)
# ---------------------------
if __name__ == '__main__':
    main()



'''
Module containing utility functions
'''
from importlib.resources import files
from functools           import cache

import yaml

# ---------------------------------
@cache
def _get_evt_name() -> dict[str,str]:
    file_path = files('ap_utilities_data').joinpath('evt_name.yaml')
    file_path = str(file_path)
    with open(file_path, encoding='utf-8') as ifile:
        d_data = yaml.safe_load(ifile)

    return d_data
# ---------------------------------
@cache
def _get_name_evt(style : str) -> dict[str,str]:
    if style != 'safe_1':
        raise NotImplementedError(f'Style {style} not supported')

    file_path = files('ap_utilities_data').joinpath('name_evt.yaml')
    file_path = str(file_path)
    with open(file_path, encoding='utf-8') as ifile:
        d_data = yaml.safe_load(ifile)

    return d_data
# ---------------------------------
def format_nickname(nickname : str, style : str) -> str:
    '''
    Function taking decays nickname and returning formatted version

    nickaname: Name to be formatted
    style    : How to format name, supported: literal, safe_1
    '''
    if style == 'literal':
        return nickname

    if style != 'safe_1':
        raise ValueError(f'Invalid style: {style}')

    nickname = nickname.replace(                  '.',     'p')
    nickname = nickname.replace(                  '-',    'mn')
    nickname = nickname.replace(                  '+',    'pl')
    nickname = nickname.replace(                  '=',  '_eq_')
    nickname = nickname.replace(                  ',',     '_')
    nickname = nickname.replace(         'DecProdCut',   'DPC')
    nickname = nickname.replace( 'EvtGenDecayWithCut', 'EGDWC')
    nickname = nickname.replace('VisibleInAcceptance',   'VIA')
    nickname = nickname.replace(        'HighVisMass',   'HVM')
    nickname = nickname.replace(       'OppositeSign',    'OS')
    nickname = nickname.replace(           'TightCut',    'TC')

    return nickname
# ---------------------------------
def read_decay_name(event_type : str, style : str = 'safe_1') -> str:
    '''
    Takes event type, and style strings, returns nickname of decay as defined in DecFiles package

    Styles:

    literal         : No change is made to nickname
    safe_1 (default): With following replacements:
        . -> p
        = -> _eq_
        - -> mn
        + -> pl
        , -> _
    '''
    d_evt_name = _get_evt_name()

    if event_type not in d_evt_name:
        raise ValueError(f'Event type {event_type} not found')

    value = d_evt_name[event_type]
    value = format_nickname(value, style)

    return value
# ---------------------------------
def read_event_type(nickname : str, style : str = 'safe_1') -> str:
    '''
    Takes nickname, and style strings, returns corresponding event type 

    Styles:

    literal         : No change is made to nickname
    safe_1 (default): With following replacements:
        . -> p
        = -> _eq_
        - -> mn
        + -> pl
        , -> _
    '''
    d_name_evt = _get_name_evt(style)

    if nickname not in d_name_evt:
        raise ValueError(f'Event type {nickname} not found')

    value = d_name_evt[nickname]

    return value
# ---------------------------------



'''
Script used to validate ntuples produced by AP pipelines
'''
import os
import glob
import shutil
import argparse
from typing              import Union
from typing              import ClassVar
from dataclasses         import dataclass
from concurrent.futures  import ThreadPoolExecutor, as_completed

import tqdm
import yaml
import pandas as pnd

from ROOT                            import TFile, TDirectoryFile, TTree
from ap_utilities.logging.log_store  import LogStore
from ap_utilities.logfiles.log_info  import LogInfo

log = LogStore.add_logger('ap_utilities_scripts:validate_ap_tuples')
# -------------------------------
@dataclass
class Data:
    '''
    Class holding shared attributes
    '''
    pipeline_id : int
    config_path : str
    nthread     : int
    cfg         : dict

    d_tree_miss      : ClassVar[dict[str, list[str]]]     = {}
    d_tree_found     : ClassVar[dict[str, list[str]]]     = {}
    d_tree_entries   : ClassVar[dict[str, dict[str,int]]] = {}
    d_mcdt           : ClassVar[dict[str, dict[str,int]]] = {}
    d_sample_entries : ClassVar[dict[str,int]]            = {}
    l_missing_job    : ClassVar[list[str]]                = []
    d_log_stat       : ClassVar[dict[str,int]]            = {}
    d_root_stat      : ClassVar[dict[str,int]]            = {}
# -------------------------------
def _check_path(path : str) -> None:
    found = os.path.isdir(path) or os.path.isfile(path)
    if not found:
        raise FileNotFoundError(f'Cannot find: {path}')
# -------------------------------
def _parse_args() -> None:
    parser = argparse.ArgumentParser(description='Makes a list of PFNs for a specific set of eventIDs in case we need to reprocess them')
    parser.add_argument('-p','--pipeline', type=int, help='Pipeline ID', required=True)
    parser.add_argument('-f','--cfg_path', type=str, help='Path to config file with the description of how to validate', required=True)
    parser.add_argument('-l','--log_lvl' , type=int, help='Logging level', default=20, choices=[10,20,30])
    parser.add_argument('-t','--nthread' , type=int, help='Number of threads', default=1)
    args = parser.parse_args()

    Data.pipeline_id = args.pipeline
    Data.config_path = args.cfg_path
    Data.nthread     = args.nthread

    LogStore.set_level('ap_utilities_scripts:validate_ap_tuples', args.log_lvl)
# -------------------------------
def _load_config() -> None:
    if not os.path.isfile(Data.config_path):
        raise FileNotFoundError(f'Could not find: {Data.config_path}')

    with open(Data.config_path, encoding='utf-8') as ifile:
        Data.cfg = yaml.safe_load(ifile)
# -------------------------------
def _get_out_paths() -> list[str]:
    '''
    Returns list of paths containing the ROOT files and the zip file with the logs
    '''
    pipeline_dir = Data.cfg['paths']['pipeline_dir']
    analysis_dir = Data.cfg['paths']['analysis_dir']

    job_path = f'{pipeline_dir}/{Data.pipeline_id}/{analysis_dir}'
    _check_path(job_path)

    sample_wc = f'{job_path}/*/*'
    l_sample  = glob.glob(sample_wc)

    nsample   = len(l_sample)
    njobs     = len(Data.cfg['samples'])

    if nsample != njobs:
        log.warning(f'Number of samples and jobs in {sample_wc} differ: {nsample} -> {njobs}')

    return l_sample
# -------------------------------
def _get_file_path(job_path : str, ending : str) -> Union[str,None]:
    path_wc = f'{job_path}/*{ending}'
    try:
        [file_path] = glob.glob(path_wc)
    except ValueError:
        log.debug(f'Cannot find one and only one file in: {path_wc}')
        return None

    return file_path
# -------------------------------
def _sample_from_path(path : str) -> str:
    '''
    Picks up path to file, returns sample associated to path
    '''
    l_samp = Data.cfg['samples']
    try:
        [samp] = [ sample for sample in l_samp if sample in path ]
    except ValueError as exc:
        raise ValueError(f'Not found one and only one sample corresponding to: {path}') from exc

    return samp
# -------------------------------
def _copy_path(source : str) -> str:
    target = f'/tmp/{source}'
    target = target.replace('#', '_')

    if os.path.isfile(target):
        return target

    target_dir = os.path.dirname(target)
    os.makedirs(target_dir, exist_ok=True)

    log.debug(f'{source} --> {target}')
    shutil.copy(source, target)

    return target
# -------------------------------
def _validate_root_file(root_path : str) -> None:
    _validate_trees(root_path)
# -------------------------------
def _add_to_dictionary(d_data : dict, identifier : str, key : str, value : int) -> None:
    if identifier not in d_data:
        d_data[identifier] = {}

    if key in d_data[identifier]:
        log.warning(f'Overriding key {key} at {identifier}')

    d_data[identifier][key] = value
# -------------------------------
def _is_valid_reco_dir(sample : str, file_dir : TDirectoryFile) -> bool:
    if hasattr(file_dir, 'MCDecayTree'):
        nentries = file_dir.MCDecayTree.GetEntries()
        _add_to_dictionary(Data.d_tree_entries, sample, key=file_dir.GetName(), value=nentries)
        return False

    if not hasattr(file_dir, 'DecayTree') or not isinstance(file_dir.DecayTree, TTree):
        _add_to_dictionary(Data.d_tree_entries, sample, key=file_dir.GetName(), value=0)
        return False

    nentries = file_dir.DecayTree.GetEntries()
    if nentries == 0:
        _add_to_dictionary(Data.d_tree_entries, sample, key=file_dir.GetName(), value=0)
        return False

    _add_to_dictionary(Data.d_tree_entries, sample, key=file_dir.GetName(), value=nentries)

    return True
# -------------------------------
def _check_mcdt_entries(sample : str, l_dir : list[TDirectoryFile]) -> dict[str,int]:
    '''
    Given a sample and a list of directories with a tree each
    If the MCDecayTree is not found return None, if it is found and the entries agree with what is in d_sample_entries
    return True, if they do not agree, return false
    '''
    l_mcdir   = [ directory for directory in l_dir if directory.GetName() == sample ]

    nexpected= Data.d_sample_entries[sample]
    if len(l_mcdir) == 0:
        return {'Expected' : nexpected, 'Found' : -1}

    tree     = l_mcdir[0].MCDecayTree
    nfound   = tree.GetEntries()

    return {'Expected' : nexpected, 'Found' : nfound}
# -------------------------------
def _validate_trees(root_path : str) -> None:
    sample    = _sample_from_path(root_path)
    l_expected= Data.cfg['samples'][sample]
    s_expected= set(l_expected)

    root_path = _copy_path(root_path)
    rfile     = TFile(root_path)
    l_key     = rfile.GetListOfKeys()
    l_dir     = [ key.ReadObj() for key in l_key if key.ReadObj().InheritsFrom('TDirectoryFile') ]
    s_found   = { fdir.GetName() for fdir in l_dir if _is_valid_reco_dir(sample, fdir)}

    Data.d_mcdt[sample] = _check_mcdt_entries(sample, l_dir)

    if s_expected == {'any'} and len(s_found) > 0:
        _save_trees(sample, s_found, Data.d_tree_found)
        return

    s_missing = s_expected - s_found

    if len(s_missing) > 0:
        log.warning(f'File: {root_path}')
        log.warning(f'Missing : {s_missing}')
        _save_trees(sample, s_missing, Data.d_tree_miss )

    _save_trees(sample, s_found, Data.d_tree_found)
# -------------------------------
def _save_trees(sample : str, s_tree_name : set[str], d_data : dict[str, list[str]]):
    l_tree_name = list(s_tree_name)
    d_data.update({sample : l_tree_name})
# -------------------------------
def _update_sample_stats(log_path : str) -> None:
    sample   = _sample_from_path(log_path)
    obj      = LogInfo(zip_path = log_path)

    Data.d_sample_entries[sample] = obj.get_mcdt_entries(sample, fall_back=-1)
# -------------------------------
def _check_job(sample : str, log_path : Union[str,None], root_path : Union[str,None]):
    if log_path is None:
        Data.d_log_stat[sample] = -1
    else:
        Data.d_log_stat[sample] = +1

    if root_path is None:
        Data.d_root_stat[sample] = -1
    else:
        Data.d_root_stat[sample] = +1
# -------------------------------
def _validate_job(job_path : str) -> None:
    '''
    Picks path to directory with ROOT and zip file
    Runs validation
    '''

    root_path = _get_file_path(job_path, ending='_2.tuple.root')
    log_path  = _get_file_path(job_path, ending=         '.zip')

    sample    = _sample_from_path(job_path)
    _check_job(sample, log_path, root_path)
    if log_path is None or root_path is None:
        Data.l_missing_job.append(job_path)
        return

    _update_sample_stats(log_path)
    _validate_root_file(root_path)
# -------------------------------
def _validate() -> None:
    _load_config()
    l_out_path = _get_out_paths()

    npath = len(l_out_path)
    log.info(f'Checking {npath} jobs')

    if Data.nthread > 1:
        _validate_with_multithreading(l_out_path)
        return

    log.info('Using single thread')
    for out_path in tqdm.tqdm(l_out_path, ascii=' -'):
        _validate_job(out_path)
# -------------------------------
def _validate_with_multithreading(l_out_path : list[str]) -> None:
    npath = len(l_out_path)
    with ThreadPoolExecutor(max_workers=Data.nthread) as executor:
        l_feat = [ executor.submit(_validate_job, out_path) for out_path in l_out_path ]

        for _ in tqdm.tqdm(as_completed(l_feat), total=npath, ascii=' -'):
            ...
# -------------------------------
def _get_mcdt_dataframe() -> pnd.DataFrame:
    l_sample   = []
    l_found    = []
    l_expected = []
    l_log_stat = []
    l_root_stat= []

    nsample = len(Data.d_mcdt)
    log.info(f'Found {nsample} samples with MCDT')

    for sample, d_stat in Data.d_mcdt.items():
        expected = d_stat['Expected']
        found    = d_stat['Found'   ]
        miss_root= Data.d_root_stat[sample]
        miss_log = Data.d_log_stat[sample]

        l_sample.append(sample)
        l_found.append(found)
        l_expected.append(expected)
        l_root_stat.append(miss_root)
        l_log_stat.append(miss_log)

    df = pnd.DataFrame({'Sample' : l_sample, 'Found' : l_found, 'Expected' : l_expected, 'Has log' : l_log_stat, 'Has tree' : l_root_stat})

    df['Difference [%]'] = 100 * (df.Expected - df.Found).abs() / df.Expected

    df = df.sort_values(by='Difference [%]', ascending=False)

    return df
# -------------------------------
def _save_report() -> None:
    d_rep = {
            'missing_trees'    : Data.d_tree_miss,
            'found_trees'      : Data.d_tree_found,
            'tree_entries'     : Data.d_tree_entries,
            'missing_jobs'     : Data.l_missing_job,
            }

    with open(f'report_{Data.pipeline_id}.yaml', 'w', encoding='utf-8') as ofile:
        yaml.safe_dump(d_rep, ofile, sort_keys=False, indent=2, default_flow_style=False)

    df = _get_mcdt_dataframe()
    with open(f'mcdt_{Data.pipeline_id}.md', 'w', encoding='utf-8') as ofile:
        df.to_markdown(ofile, index=False)
# -------------------------------
def main():
    '''
    Script starts here
    '''
    _parse_args()
    _validate()
    _save_report()
# -------------------------------
if __name__ == '__main__':
    main()



'''
Module with functions used to test functions in physics/utilities.py
'''
import pytest

import ap_utilities.decays.utilities as aput

# --------------------------------------------------
class Data:
    '''
    Class used to store data needed by tests
    '''

    l_event_type = [
        '10000000',
        '10000010',
        '10000020',
        '10000021',
        '10000022',
        '10000023',
        '10000027',
        '10000030',
        '10002203',
        '10002213',
        '11100001',
        '11100003',
        '11100006',
        ]

    l_nickname = [
        'tau_5mu_eq_DPC',
        'tau_Kphinu_KK_eq_DPC',
        'tau_mugamma_eq_DPC',
        'tau_mumnKplKmn_eq_DPC',
        'tau_mumnpiplpimn_eq_DPC',
        'tau_mumue_eq_OS_DPC',
        'tau_mumue_eq_OS_FromB_TightCut',
        'tau_mumue_eq_SS_DPC',
        'tau_mumue_eq_SS_FromB_TightCut',
        'tau_mumumu_eq_DPC',
        'tau_mumumu_eq_FromB',
        'tau_muphi_KK_eq_DPC',
        'tau_muphi_KK_eq_FromB',
        'tau_muplpimnpimn_eq_DPC',
        'tau_nupiplpi0_eegamma_eq_DPC',
        'tau_pimnpiplpimnnu_eq_DPC',
        'tau_pimnpiplpimnpi0nu_eq_DPC',
        'tau_piphinu_KK_eq_DPC',
            ]
# --------------------------------------------------
@pytest.mark.parametrize('event_type', Data.l_event_type)
def test_read_decay_name(event_type : str) -> None:
    '''
    Tests reading of decay name from YAML using event type
    '''
    literal = aput.read_decay_name(event_type=event_type, style='literal')
    safe_1  = aput.read_decay_name(event_type=event_type, style= 'safe_1')

    print(f'{literal:<50}{safe_1:<50}')
# --------------------------------------------------
@pytest.mark.parametrize('nickname', Data.l_nickname)
def test_read_event_type(nickname: str) -> None:
    '''
    Tests reading of event type from YAML using nickname 
    '''
    event_type = aput.read_event_type(nickname=nickname, style= 'safe_1')
    print(event_type)
# --------------------------------------------------
